/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { IFluidHandle } from "@fluidframework/core-interfaces";
import { IForestSubscription, TreeValue, UpPath } from "../core/index.js";
import { FlexAllowedTypes, FlexObjectNodeSchema, FlexTreeField, FlexTreeNode, FlexTreeNodeSchema, FlexTreeSequenceField } from "../feature-libraries/index.js";
import { TreeMapNode, type TreeNodeSchema } from "./schemaTypes.js";
import { TreeArrayNode } from "./treeArrayNode.js";
import { TreeNode, Unhydrated } from "./types.js";
/**
 * Detects if the given 'candidate' is a TreeNode.
 *
 * @remarks
 * Supports both Hydrated and {@link Unhydrated} TreeNodes, both of which return true.
 *
 * Because the common usage is to check if a value being inserted/set is a TreeNode,
 * this function permits calling with primitives as well as objects.
 *
 * Primitives will always return false (as they are copies of data, not references to nodes).
 *
 * @param candidate - Value which may be a TreeNode
 * @returns true if the given 'candidate' is a hydrated TreeNode.
 */
export declare function isTreeNode(candidate: unknown): candidate is TreeNode | Unhydrated<TreeNode>;
/**
 * Retrieve the associated proxy for the given field.
 * */
export declare function getProxyForField(field: FlexTreeField): TreeNode | TreeValue | undefined;
/**
 * A symbol for storing TreeNodeSchema on FlexTreeNode's schema.
 */
export declare const simpleSchemaSymbol: unique symbol;
export declare function getSimpleSchema(schema: FlexTreeNodeSchema): TreeNodeSchema | undefined;
export declare function getOrCreateNodeProxy(flexNode: FlexTreeNode): TreeNode | TreeValue;
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `{}` is used for the target.
 */
export declare function createObjectProxy<TSchema extends FlexObjectNodeSchema>(schema: TSchema, allowAdditionalProperties: boolean, targetObject?: object): TreeNode;
/**
 * Given a array node proxy, returns its underlying LazySequence field.
 */
export declare const getSequenceField: <TTypes extends FlexAllowedTypes>(arrayNode: TreeArrayNode) => FlexTreeSequenceField<TTypes>;
/**
 * PropertyDescriptorMap used to build the prototype for our array node dispatch object.
 */
export declare const arrayNodePrototypeProperties: PropertyDescriptorMap;
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `[]` is used for the target and a separate object created to dispatch array methods.
 * If provided, the customTargetObject will be used as both the dispatch object and the proxy target, and therefor must provide an own `length` value property
 * (which is not used but must exist for getOwnPropertyDescriptor invariants) and the array functionality from {@link arrayNodePrototype}.
 */
export declare function createArrayNodeProxy(allowAdditionalProperties: boolean, customTargetObject?: object): TreeArrayNode;
export declare const mapStaticDispatchMap: PropertyDescriptorMap;
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `new Map()` is used for the target and a separate object created to dispatch map methods.
 * If provided, the customTargetObject will be used as both the dispatch object and the proxy target, and therefor must provide the map functionality from {@link mapPrototype}.
 */
export declare function createMapProxy(allowAdditionalProperties: boolean, customTargetObject?: object): TreeMapNode;
/**
 * Records any proxies in the given content tree and does the necessary bookkeeping to ensure they are synchronized with subsequent reads of the tree.
 * @remarks If the content tree contains any proxies, this function must be called just prior to inserting the content into the tree.
 * Specifically, no other content may be inserted into the tree between the invocation of this function and the insertion of `content`.
 * The insertion of `content` must occur or else this function will cause memory leaks.
 * @param content - the tree of content to be inserted, of which any of its object/map/array nodes might be a proxy
 * @param anchors - the {@link AnchorSet} for the tree
 * @returns The content after having all proxies replaced inline with plain javascript objects.
 * See {@link extractFactoryContent} for more details.
 */
export declare function prepareContentForInsert(content: InsertableContent, forest: IForestSubscription): FactoryContent;
/**
 * Given a content tree that is to be inserted into the shared tree, replace all subtrees that were created by factories
 * (via {@link SharedTreeObjectFactory.create}) with the content that was passed to those factories.
 * @param content - the content being inserted which may be, and/or may contain, factory-created content
 * @param onVisitProxy - an optional callback that will run for each proxy (i.e. object created by a factory) found in the inserted content
 * @param insertedAtIndex - if the content being inserted is array node content, this must be the index in the array node at which the content is being inserted
 * @returns the result of the content replacement and a {@link ExtractedFactoryContent.hydrateProxies} function which must be invoked if present.
 * @remarks
 * This functions works recursively.
 * Factory-created objects that are nested inside of the content passed to other factory-created objects, and so on, will be in-lined.
 * This function also adds the hidden {@link typeNameSymbol} of each object schema to the output.
 * @example
 * ```ts
 * const x = foo.create({
 *   a: 3, b: bar.create({
 *     c: [baz.create({ d: 5 })]
 *   })
 * });
 * const y = extractFactoryContent(x);
 * y === {
 *   [typeNameSymbol]: "foo", a: 3, b: {
 *     [typeNameSymbol]: "bar", c: [{ [typeNameSymbol]: "baz", d: 5 }]
 *  }
 * }
 * ```
 */
export declare function extractFactoryContent(input: InsertableContent, visitProxies?: {
    path: UpPath;
    onVisitProxy: (path: UpPath, proxy: TreeNode) => void;
}): FactoryContent;
/**
 * Content which can be used to build a node.
 * @remarks
 * Can contain unhydrated nodes, but can not be an unhydrated node at the root.
 */
export type FactoryContent = IFluidHandle | string | number | boolean | null | ReadonlyMap<string, InsertableContent> | readonly InsertableContent[] | {
    readonly [P in string]?: InsertableContent;
};
/**
 * Content which can be inserted into a tree.
 */
export type InsertableContent = Unhydrated<TreeNode> | FactoryContent;
//# sourceMappingURL=proxies.d.ts.map