"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractFactoryContent = exports.prepareContentForInsert = exports.createMapProxy = exports.mapStaticDispatchMap = exports.createArrayNodeProxy = exports.arrayNodePrototypeProperties = exports.getSequenceField = exports.createObjectProxy = exports.getOrCreateNodeProxy = exports.getSimpleSchema = exports.simpleSchemaSymbol = exports.getProxyForField = exports.isTreeNode = void 0;
const core_utils_1 = require("@fluidframework/core-utils");
const index_js_1 = require("../core/index.js");
// TODO: decide how to deal with dependencies on flex-tree implementation.
// eslint-disable-next-line import/no-internal-modules
const lazyNode_js_1 = require("../feature-libraries/flex-tree/lazyNode.js");
const index_js_2 = require("../feature-libraries/index.js");
const index_js_3 = require("../util/index.js");
const proxyBinding_js_1 = require("./proxyBinding.js");
const rawNode_js_1 = require("./rawNode.js");
const schemaTypes_js_1 = require("./schemaTypes.js");
const toMapTree_js_1 = require("./toMapTree.js");
const treeArrayNode_js_1 = require("./treeArrayNode.js");
/**
 * Detects if the given 'candidate' is a TreeNode.
 *
 * @remarks
 * Supports both Hydrated and {@link Unhydrated} TreeNodes, both of which return true.
 *
 * Because the common usage is to check if a value being inserted/set is a TreeNode,
 * this function permits calling with primitives as well as objects.
 *
 * Primitives will always return false (as they are copies of data, not references to nodes).
 *
 * @param candidate - Value which may be a TreeNode
 * @returns true if the given 'candidate' is a hydrated TreeNode.
 */
function isTreeNode(candidate) {
    return (0, proxyBinding_js_1.tryGetFlexNode)(candidate) !== undefined;
}
exports.isTreeNode = isTreeNode;
/**
 * Retrieve the associated proxy for the given field.
 * */
function getProxyForField(field) {
    switch (field.schema.kind) {
        case index_js_2.FieldKinds.required: {
            const asValue = field;
            // TODO: Ideally, we would return leaves without first boxing them.  However, this is not
            //       as simple as calling '.content' since this skips the node and returns the FieldNode's
            //       inner field.
            return getOrCreateNodeProxy(asValue.boxedContent);
        }
        case index_js_2.FieldKinds.optional: {
            const asValue = field;
            // TODO: Ideally, we would return leaves without first boxing them.  However, this is not
            //       as simple as calling '.content' since this skips the node and returns the FieldNode's
            //       inner field.
            const maybeContent = asValue.boxedContent;
            // Normally, empty fields are unreachable due to the behavior of 'tryGetField'.  However, the
            // root field is a special case where the field is always present (even if empty).
            return maybeContent === undefined ? undefined : getOrCreateNodeProxy(maybeContent);
        }
        // TODO: Remove if/when 'FieldNode' is removed.
        case index_js_2.FieldKinds.sequence: {
            // 'getProxyForNode' handles FieldNodes by unconditionally creating a array node proxy, making
            // this case unreachable as long as users follow the 'array recipe'.
            (0, index_js_3.fail)("'sequence' field is unexpected.");
        }
        default:
            (0, index_js_3.fail)("invalid field kind");
    }
}
exports.getProxyForField = getProxyForField;
/**
 * A symbol for storing TreeNodeSchema on FlexTreeNode's schema.
 */
exports.simpleSchemaSymbol = Symbol(`simpleSchema`);
function getSimpleSchema(schema) {
    if (exports.simpleSchemaSymbol in schema) {
        return schema[exports.simpleSchemaSymbol];
    }
    return undefined;
}
exports.getSimpleSchema = getSimpleSchema;
function getOrCreateNodeProxy(flexNode) {
    const cachedProxy = (0, proxyBinding_js_1.tryGetProxy)(flexNode);
    if (cachedProxy !== undefined) {
        return cachedProxy;
    }
    const schema = flexNode.schema;
    const classSchema = getSimpleSchema(schema);
    (0, core_utils_1.assert)(classSchema !== undefined, "node without schema");
    if (typeof classSchema === "function") {
        const simpleSchema = classSchema;
        return new simpleSchema(flexNode);
    }
    else {
        return classSchema.create(flexNode);
    }
}
exports.getOrCreateNodeProxy = getOrCreateNodeProxy;
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `{}` is used for the target.
 */
function createObjectProxy(schema, allowAdditionalProperties, targetObject = {}) {
    // To satisfy 'deepEquals' level scrutiny, the target of the proxy must be an object with the same
    // prototype as an object literal '{}'.  This is because 'deepEquals' uses 'Object.getPrototypeOf'
    // as a way to quickly reject objects with different prototype chains.
    //
    // (Note that the prototype of an object literal appears as '[Object: null prototype] {}', not because
    // the prototype is null, but because the prototype object itself has a null prototype.)
    // TODO: Although the target is an object literal, it's still worthwhile to try experimenting with
    // a dispatch object to see if it improves performance.
    const proxy = new Proxy(targetObject, {
        get(target, key) {
            const field = (0, proxyBinding_js_1.getFlexNode)(proxy).tryGetField(key);
            if (field !== undefined) {
                return getProxyForField(field);
            }
            // Pass the proxy as the receiver here, so that any methods on the prototype receive `proxy` as `this`.
            return Reflect.get(target, key, proxy);
        },
        set(target, key, value) {
            const flexNode = (0, proxyBinding_js_1.getFlexNode)(proxy);
            const flexNodeSchema = flexNode.schema;
            (0, core_utils_1.assert)(flexNodeSchema instanceof index_js_2.FlexObjectNodeSchema, 0x888 /* invalid schema */);
            const fieldSchema = flexNodeSchema.objectNodeFields.get(key);
            if (fieldSchema === undefined) {
                return allowAdditionalProperties ? Reflect.set(target, key, value) : false;
            }
            // TODO: Is it safe to assume 'content' is a LazyObjectNode?
            (0, core_utils_1.assert)(flexNode instanceof lazyNode_js_1.LazyObjectNode, 0x7e0 /* invalid content */);
            (0, core_utils_1.assert)(typeof key === "string", 0x7e1 /* invalid key */);
            const field = (0, lazyNode_js_1.getBoxedField)(flexNode, (0, index_js_3.brand)(key), fieldSchema);
            switch (field.schema.kind) {
                case index_js_2.FieldKinds.required:
                case index_js_2.FieldKinds.optional: {
                    const typedField = field;
                    const content = prepareContentForInsert(value, flexNode.context.forest);
                    const cursor = (0, toMapTree_js_1.cursorFromNodeData)(content, flexNode.context.schema, fieldSchema.allowedTypeSet);
                    typedField.content = cursor;
                    break;
                }
                default:
                    (0, index_js_3.fail)("invalid FieldKind");
            }
            return true;
        },
        has: (target, key) => {
            return (schema.objectNodeFields.has(key) ||
                (allowAdditionalProperties ? Reflect.has(target, key) : false));
        },
        ownKeys: (target) => {
            return [
                ...schema.objectNodeFields.keys(),
                ...(allowAdditionalProperties ? Reflect.ownKeys(target) : []),
            ];
        },
        getOwnPropertyDescriptor: (target, key) => {
            const field = (0, proxyBinding_js_1.getFlexNode)(proxy).tryGetField(key);
            if (field === undefined) {
                return allowAdditionalProperties
                    ? Reflect.getOwnPropertyDescriptor(target, key)
                    : undefined;
            }
            const p = {
                value: getProxyForField(field),
                writable: true,
                enumerable: true,
                configurable: true, // Must be 'configurable' if property is absent from proxy target.
            };
            return p;
        },
    });
    return proxy;
}
exports.createObjectProxy = createObjectProxy;
/**
 * Given a array node proxy, returns its underlying LazySequence field.
 */
const getSequenceField = (arrayNode) => (0, proxyBinding_js_1.getFlexNode)(arrayNode).content;
exports.getSequenceField = getSequenceField;
// Used by 'insert*()' APIs to converts new content (expressed as a proxy union) to contextually
// typed data prior to forwarding to 'LazySequence.insert*()'.
function contextualizeInsertedArrayContent(content, sequenceField) {
    return prepareContentForInsert(content.flatMap((c) => c instanceof treeArrayNode_js_1.IterableTreeArrayContent ? Array.from(c) : [c]), sequenceField.context.forest);
}
// #region Create dispatch map for array nodes
// TODO: Experiment with alternative dispatch methods to see if we can improve performance.
/**
 * PropertyDescriptorMap used to build the prototype for our array node dispatch object.
 */
exports.arrayNodePrototypeProperties = {
    // We manually add [Symbol.iterator] to the dispatch map rather than use '[fn.name] = fn' as
    // below when adding 'Array.prototype.*' properties to this map because 'Array.prototype[Symbol.iterator].name'
    // returns "values" (i.e., Symbol.iterator is an alias for the '.values()' function.)
    [Symbol.iterator]: {
        value: Array.prototype[Symbol.iterator],
    },
    at: {
        value(index) {
            const field = (0, exports.getSequenceField)(this);
            const val = field.boxedAt(index);
            if (val === undefined) {
                return val;
            }
            return getOrCreateNodeProxy(val);
        },
    },
    insertAt: {
        value(index, ...value) {
            const sequenceField = (0, exports.getSequenceField)(this);
            const content = contextualizeInsertedArrayContent(value, sequenceField);
            sequenceField.insertAt(index, (0, toMapTree_js_1.cursorFromFieldData)(content, sequenceField.context.schema, sequenceField.schema));
        },
    },
    insertAtStart: {
        value(...value) {
            const sequenceField = (0, exports.getSequenceField)(this);
            const content = contextualizeInsertedArrayContent(value, sequenceField);
            sequenceField.insertAtStart((0, toMapTree_js_1.cursorFromFieldData)(content, sequenceField.context.schema, sequenceField.schema));
        },
    },
    insertAtEnd: {
        value(...value) {
            const sequenceField = (0, exports.getSequenceField)(this);
            const content = contextualizeInsertedArrayContent(value, sequenceField);
            sequenceField.insertAtEnd((0, toMapTree_js_1.cursorFromFieldData)(content, sequenceField.context.schema, sequenceField.schema));
        },
    },
    removeAt: {
        value(index) {
            (0, exports.getSequenceField)(this).removeAt(index);
        },
    },
    removeRange: {
        value(start, end) {
            (0, exports.getSequenceField)(this).removeRange(start, end);
        },
    },
    moveToStart: {
        value(sourceIndex, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveToStart(sourceIndex, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveToStart(sourceIndex);
            }
        },
    },
    moveToEnd: {
        value(sourceIndex, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveToEnd(sourceIndex, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveToEnd(sourceIndex);
            }
        },
    },
    moveToIndex: {
        value(index, sourceIndex, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveToIndex(index, sourceIndex, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveToIndex(index, sourceIndex);
            }
        },
    },
    moveRangeToStart: {
        value(sourceStart, sourceEnd, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveRangeToStart(sourceStart, sourceEnd, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveRangeToStart(sourceStart, sourceEnd);
            }
        },
    },
    moveRangeToEnd: {
        value(sourceStart, sourceEnd, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveRangeToEnd(sourceStart, sourceEnd, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveRangeToEnd(sourceStart, sourceEnd);
            }
        },
    },
    moveRangeToIndex: {
        value(index, sourceStart, sourceEnd, source) {
            if (source !== undefined) {
                (0, exports.getSequenceField)(this).moveRangeToIndex(index, sourceStart, sourceEnd, (0, exports.getSequenceField)(source));
            }
            else {
                (0, exports.getSequenceField)(this).moveRangeToIndex(index, sourceStart, sourceEnd);
            }
        },
    },
};
/* eslint-disable @typescript-eslint/unbound-method */
// For compatibility, we are initially implement 'readonly T[]' by applying the Array.prototype methods
// to the array node proxy.  Over time, we should replace these with efficient implementations on LazySequence
// to avoid re-entering the proxy as these methods access 'length' and the indexed properties.
//
// For brevity, the current implementation dynamically builds a property descriptor map from a list of
// Array functions we want to re-expose via the proxy.
// TODO: This assumes 'Function.name' matches the property name on 'Array.prototype', which may be
// dubious across JS engines.
[
    Array.prototype.concat,
    // Array.prototype.copyWithin,
    Array.prototype.entries,
    Array.prototype.every,
    // Array.prototype.fill,
    Array.prototype.filter,
    Array.prototype.find,
    Array.prototype.findIndex,
    Array.prototype.flat,
    Array.prototype.flatMap,
    Array.prototype.forEach,
    Array.prototype.includes,
    Array.prototype.indexOf,
    Array.prototype.join,
    Array.prototype.keys,
    Array.prototype.lastIndexOf,
    // Array.prototype.length,
    Array.prototype.map,
    // Array.prototype.pop,
    // Array.prototype.push,
    Array.prototype.reduce,
    Array.prototype.reduceRight,
    // Array.prototype.reverse,
    // Array.prototype.shift,
    Array.prototype.slice,
    Array.prototype.some,
    // Array.prototype.sort,
    // Array.prototype.splice,
    Array.prototype.toLocaleString,
    Array.prototype.toString,
    // Array.prototype.unshift,
    Array.prototype.values,
].forEach((fn) => {
    exports.arrayNodePrototypeProperties[fn.name] = { value: fn };
});
/* eslint-enable @typescript-eslint/unbound-method */
const arrayNodePrototype = Object.create(Object.prototype, exports.arrayNodePrototypeProperties);
// #endregion
/**
 * Helper to coerce property keys to integer indexes (or undefined if not an in-range integer).
 */
function asIndex(key, length) {
    if (typeof key === "string") {
        // TODO: It may be worth a '0' <= ch <= '9' check before calling 'Number' to quickly
        // reject 'length' as an index, or even parsing integers ourselves.
        const asNumber = Number(key);
        // TODO: See 'matrix/range.ts' for fast integer coercing + range check.
        if (Number.isInteger(asNumber)) {
            return 0 <= asNumber && asNumber < length ? asNumber : undefined;
        }
    }
}
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `[]` is used for the target and a separate object created to dispatch array methods.
 * If provided, the customTargetObject will be used as both the dispatch object and the proxy target, and therefor must provide an own `length` value property
 * (which is not used but must exist for getOwnPropertyDescriptor invariants) and the array functionality from {@link arrayNodePrototype}.
 */
function createArrayNodeProxy(allowAdditionalProperties, customTargetObject) {
    const targetObject = customTargetObject ?? [];
    // Create a 'dispatch' object that this Proxy forwards to instead of the proxy target, because we need
    // the proxy target to be a plain JS array (see comments below when we instantiate the Proxy).
    // Own properties on the dispatch object are surfaced as own properties of the proxy.
    // (e.g., 'length', which is defined below).
    //
    // Properties normally inherited from 'Array.prototype' are surfaced via the prototype chain.
    const dispatch = customTargetObject ??
        Object.create(arrayNodePrototype, {
            // This dispatch object's set of keys is used to implement `has` (for the `in` operator) for the non-numeric cases, and therefor must include `length`.
            length: {
                get() {
                    (0, index_js_3.fail)("Proxy should intercept length");
                },
                set() {
                    (0, index_js_3.fail)("Proxy should intercept length");
                },
                enumerable: false,
                configurable: false,
            },
        });
    // To satisfy 'deepEquals' level scrutiny, the target of the proxy must be an array literal in order
    // to pass 'Object.getPrototypeOf'.  It also satisfies 'Array.isArray' and 'Object.prototype.toString'
    // requirements without use of Array[Symbol.species], which is potentially on a path ot deprecation.
    const proxy = new Proxy(targetObject, {
        get: (target, key) => {
            const field = (0, exports.getSequenceField)(proxy);
            const maybeIndex = asIndex(key, field.length);
            if (maybeIndex === undefined) {
                if (key === "length") {
                    return field.length;
                }
                // Pass the proxy as the receiver here, so that any methods on
                // the prototype receive `proxy` as `this`.
                return Reflect.get(dispatch, key, proxy);
            }
            const value = field.boxedAt(maybeIndex);
            if (value === undefined) {
                return undefined;
            }
            // TODO: Ideally, we would return leaves without first boxing them.  However, this is not
            //       as simple as calling '.content' since this skips the node and returns the FieldNode's
            //       inner field.
            return getOrCreateNodeProxy(value);
        },
        set: (target, key, newValue, receiver) => {
            if (key === "length") {
                // To allow "length" to look like "length" on an array, getOwnPropertyDescriptor has to report it as a writable value.
                // This means the proxy target must provide a length value, but since it can't use getters and setters, it can't be correct.
                // Therefor length has to be handled in this proxy.
                // Since its not actually mutable, return false so setting it will produce a type error.
                return false;
            }
            // 'Symbol.isConcatSpreadable' may be set on an Array instance to modify the behavior of
            // the concat method.  We allow this property to be added to the dispatch object.
            if (key === Symbol.isConcatSpreadable) {
                return Reflect.set(dispatch, key, newValue, proxy);
            }
            const field = (0, exports.getSequenceField)(proxy);
            const maybeIndex = asIndex(key, field.length);
            if (maybeIndex !== undefined) {
                // For MVP, we otherwise disallow setting properties (mutation is only available via the array node mutation APIs).
                return false;
            }
            return allowAdditionalProperties ? Reflect.set(target, key, newValue) : false;
        },
        has: (target, key) => {
            const field = (0, exports.getSequenceField)(proxy);
            const maybeIndex = asIndex(key, field.length);
            return maybeIndex !== undefined || Reflect.has(dispatch, key);
        },
        ownKeys: (target) => {
            const field = (0, exports.getSequenceField)(proxy);
            // TODO: Would a lazy iterator to produce the indexes work / be more efficient?
            // TODO: Need to surface 'Symbol.isConcatSpreadable' as an own key.
            const keys = Array.from({ length: field.length }, (_, index) => `${index}`);
            if (allowAdditionalProperties) {
                keys.push(...Reflect.ownKeys(target));
            }
            else {
                keys.push("length");
            }
            return keys;
        },
        getOwnPropertyDescriptor: (target, key) => {
            const field = (0, exports.getSequenceField)(proxy);
            const maybeIndex = asIndex(key, field.length);
            if (maybeIndex !== undefined) {
                const val = field.boxedAt(maybeIndex);
                // To satisfy 'deepEquals' level scrutiny, the property descriptor for indexed properties must
                // be a simple value property (as opposed to using getter) and declared writable/enumerable/configurable.
                return {
                    // TODO: Ideally, we would return leaves without first boxing them.  However, this is not
                    //       as simple as calling '.at' since this skips the node and returns the FieldNode's
                    //       inner field.
                    value: val === undefined ? val : getOrCreateNodeProxy(val),
                    writable: true,
                    enumerable: true,
                    configurable: true,
                };
            }
            else if (key === "length") {
                // To satisfy 'deepEquals' level scrutiny, the property descriptor for 'length' must be a simple
                // value property (as opposed to using getter) and be declared writable / non-configurable.
                return {
                    value: field.length,
                    writable: true,
                    enumerable: false,
                    configurable: false,
                };
            }
            return Reflect.getOwnPropertyDescriptor(dispatch, key);
        },
    });
    return proxy;
}
exports.createArrayNodeProxy = createArrayNodeProxy;
// #region Create dispatch map for maps
exports.mapStaticDispatchMap = {
    [Symbol.iterator]: {
        value() {
            return this.entries();
        },
    },
    delete: {
        value(key) {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            node.delete(key);
        },
    },
    entries: {
        *value() {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            for (const key of node.keys()) {
                yield [key, getProxyForField(node.getBoxed(key))];
            }
        },
    },
    get: {
        value(key) {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            const field = node.getBoxed(key);
            return getProxyForField(field);
        },
    },
    has: {
        value(key) {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            return node.has(key);
        },
    },
    keys: {
        value() {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            return node.keys();
        },
    },
    set: {
        value(key, value) {
            const node = (0, proxyBinding_js_1.getFlexNode)(this);
            const content = prepareContentForInsert(value, node.context.forest);
            const cursor = (0, toMapTree_js_1.cursorFromNodeData)(content, node.context.schema, node.schema.mapFields.allowedTypeSet);
            node.set(key, cursor);
            return this;
        },
    },
    size: {
        get() {
            return (0, proxyBinding_js_1.getFlexNode)(this).size;
        },
    },
    values: {
        *value() {
            for (const [, value] of this.entries()) {
                yield value;
            }
        },
    },
    forEach: {
        value(callbackFn, thisArg) {
            if (thisArg === undefined) {
                // We can't pass `callbackFn` to `FlexTreeMapNode` directly, or else the third argument ("map") will be a flex node instead of the proxy.
                (0, proxyBinding_js_1.getFlexNode)(this).forEach((v, k, _) => callbackFn(v, k, this), thisArg);
            }
            else {
                const boundCallbackFn = callbackFn.bind(thisArg);
                (0, proxyBinding_js_1.getFlexNode)(this).forEach((v, k, _) => boundCallbackFn(v, k, this), thisArg);
            }
        },
    },
    // TODO: add `clear` once we have established merge semantics for it.
};
const mapPrototype = Object.create(Object.prototype, exports.mapStaticDispatchMap);
// #endregion
/**
 * @param allowAdditionalProperties - If true, setting of unexpected properties will be forwarded to the target object.
 * Otherwise setting of unexpected properties will error.
 * @param customTargetObject - Target object of the proxy.
 * If not provided `new Map()` is used for the target and a separate object created to dispatch map methods.
 * If provided, the customTargetObject will be used as both the dispatch object and the proxy target, and therefor must provide the map functionality from {@link mapPrototype}.
 */
function createMapProxy(allowAdditionalProperties, customTargetObject) {
    // Create a 'dispatch' object that this Proxy forwards to instead of the proxy target.
    const dispatch = customTargetObject ??
        Object.create(mapPrototype, {
        // Empty - JavaScript Maps do not expose any "own" properties.
        });
    const targetObject = customTargetObject ?? new Map();
    // TODO: Although the target is an object literal, it's still worthwhile to try experimenting with
    // a dispatch object to see if it improves performance.
    const proxy = new Proxy(targetObject, {
        get: (target, key, receiver) => {
            // Pass the proxy as the receiver here, so that any methods on the prototype receive `proxy` as `this`.
            return Reflect.get(dispatch, key, proxy);
        },
        getOwnPropertyDescriptor: (target, key) => {
            return Reflect.getOwnPropertyDescriptor(dispatch, key);
        },
        has: (target, key) => {
            return Reflect.has(dispatch, key);
        },
        set: (target, key, newValue) => {
            return allowAdditionalProperties ? Reflect.set(dispatch, key, newValue) : false;
        },
        ownKeys: (target) => {
            // All of Map's properties are inherited via its prototype, so there is nothing to return here,
            return [];
        },
    });
    return proxy;
}
exports.createMapProxy = createMapProxy;
/**
 * Records any proxies in the given content tree and does the necessary bookkeeping to ensure they are synchronized with subsequent reads of the tree.
 * @remarks If the content tree contains any proxies, this function must be called just prior to inserting the content into the tree.
 * Specifically, no other content may be inserted into the tree between the invocation of this function and the insertion of `content`.
 * The insertion of `content` must occur or else this function will cause memory leaks.
 * @param content - the tree of content to be inserted, of which any of its object/map/array nodes might be a proxy
 * @param anchors - the {@link AnchorSet} for the tree
 * @returns The content after having all proxies replaced inline with plain javascript objects.
 * See {@link extractFactoryContent} for more details.
 */
function prepareContentForInsert(content, forest) {
    if ((0, index_js_3.isReadonlyArray)(content)) {
        return prepareArrayContentForInsert(content, forest);
    }
    const proxies = {
        rootPath: { parent: undefined, parentField: index_js_1.EmptyKey, parentIndex: 0 },
        proxyPaths: [],
    };
    const extractedContent = extractFactoryContent(content, {
        path: proxies.rootPath,
        onVisitProxy: (p, proxy) => {
            proxies.proxyPaths.push({ path: p, proxy });
        },
    });
    bindProxies([proxies], forest);
    return extractedContent;
}
exports.prepareContentForInsert = prepareContentForInsert;
function prepareArrayContentForInsert(content, forest) {
    const proxies = [];
    const extractedContent = [];
    for (let i = 0; i < content.length; i++) {
        proxies.push({
            rootPath: {
                parent: undefined,
                parentField: index_js_1.EmptyKey,
                parentIndex: 0,
            },
            proxyPaths: [],
        });
        extractedContent.push(extractFactoryContent(content[i], {
            path: proxies[i].rootPath,
            onVisitProxy: (p, proxy) => {
                proxies[i].proxyPaths.push({ path: p, proxy });
            },
        }));
    }
    bindProxies(proxies, forest);
    return extractedContent;
}
function bindProxies(proxies, forest) {
    // Only subscribe to the event if there is at least one proxy tree to hydrate - this is not the case when inserting an empty array [].
    if (proxies.length > 0) {
        // Creating a new array emits one event per element in the array, so listen to the event once for each element
        let i = 0;
        const off = forest.on("afterRootFieldCreated", (fieldKey) => {
            proxies[i].rootPath.parentField = fieldKey;
            for (const { path, proxy } of proxies[i].proxyPaths) {
                (0, proxyBinding_js_1.anchorProxy)(forest.anchors, path, proxy);
            }
            if (++i === proxies.length) {
                off();
            }
        });
    }
}
/**
 * Given a content tree that is to be inserted into the shared tree, replace all subtrees that were created by factories
 * (via {@link SharedTreeObjectFactory.create}) with the content that was passed to those factories.
 * @param content - the content being inserted which may be, and/or may contain, factory-created content
 * @param onVisitProxy - an optional callback that will run for each proxy (i.e. object created by a factory) found in the inserted content
 * @param insertedAtIndex - if the content being inserted is array node content, this must be the index in the array node at which the content is being inserted
 * @returns the result of the content replacement and a {@link ExtractedFactoryContent.hydrateProxies} function which must be invoked if present.
 * @remarks
 * This functions works recursively.
 * Factory-created objects that are nested inside of the content passed to other factory-created objects, and so on, will be in-lined.
 * This function also adds the hidden {@link typeNameSymbol} of each object schema to the output.
 * @example
 * ```ts
 * const x = foo.create({
 *   a: 3, b: bar.create({
 *     c: [baz.create({ d: 5 })]
 *   })
 * });
 * const y = extractFactoryContent(x);
 * y === {
 *   [typeNameSymbol]: "foo", a: 3, b: {
 *     [typeNameSymbol]: "bar", c: [{ [typeNameSymbol]: "baz", d: 5 }]
 *  }
 * }
 * ```
 */
function extractFactoryContent(input, visitProxies) {
    let content;
    const rawFlexNode = (0, proxyBinding_js_1.tryGetFlexNode)(input);
    if (rawFlexNode !== undefined) {
        const factoryContent = (0, rawNode_js_1.extractRawNodeContent)(rawFlexNode);
        if (factoryContent === undefined) {
            // We were passed a proxy, but that proxy doesn't have any raw content.
            throw new Error("Cannot insert a node that is already in the tree");
        }
        visitProxies?.onVisitProxy(visitProxies.path, input);
        content = factoryContent;
    }
    else {
        content = input;
    }
    (0, core_utils_1.assert)(!isTreeNode(content), 0x844 /* Unhydrated insertion content should have FlexNode */);
    let type;
    let extractedContent;
    if ((0, index_js_3.isReadonlyArray)(content)) {
        type = schemaTypes_js_1.NodeKind.Array;
        extractedContent = extractContentArray(content, visitProxies);
    }
    else if (content instanceof Map) {
        type = schemaTypes_js_1.NodeKind.Map;
        extractedContent = extractContentMap(content, visitProxies);
    }
    else if (typeof content === "object" && content !== null && !(0, index_js_2.isFluidHandle)(content)) {
        type = schemaTypes_js_1.NodeKind.Object;
        extractedContent = extractContentObject(content, visitProxies);
    }
    else {
        extractedContent = content;
        type = schemaTypes_js_1.NodeKind.Leaf;
    }
    if (rawFlexNode !== undefined) {
        const kindFromSchema = getSimpleSchema(rawFlexNode.schema)?.kind ??
            (0, index_js_3.fail)("NodeBase should always have class schema");
        (0, core_utils_1.assert)(kindFromSchema === type, 0x845 /* kind of data should match kind of schema */);
    }
    return extractedContent;
}
exports.extractFactoryContent = extractFactoryContent;
function extractContentArray(input, visitProxies) {
    const output = [];
    if (index_js_2.typeNameSymbol in input) {
        Object.defineProperty(output, index_js_2.typeNameSymbol, { value: input[index_js_2.typeNameSymbol] });
    }
    for (let i = 0; i < input.length; i++) {
        const childContent = extractFactoryContent(input[i], visitProxies !== undefined
            ? {
                path: {
                    parent: visitProxies.path,
                    parentField: index_js_1.EmptyKey,
                    parentIndex: i,
                },
                onVisitProxy: visitProxies?.onVisitProxy,
            }
            : undefined);
        output.push(childContent);
    }
    return output;
}
function extractContentMap(input, visitProxies) {
    const output = new Map();
    if (index_js_2.typeNameSymbol in input) {
        Object.defineProperty(output, index_js_2.typeNameSymbol, { value: input[index_js_2.typeNameSymbol] });
    }
    for (const [key, value] of input) {
        const childContent = extractFactoryContent(value, visitProxies !== undefined
            ? {
                path: {
                    parent: visitProxies.path,
                    parentField: (0, index_js_3.brand)(key),
                    parentIndex: 0,
                },
                onVisitProxy: visitProxies?.onVisitProxy,
            }
            : undefined);
        output.set(key, childContent);
    }
    return output;
}
function extractContentObject(input, visitProxies) {
    const output = {};
    if (index_js_2.typeNameSymbol in input) {
        Object.defineProperty(output, index_js_2.typeNameSymbol, { value: input[index_js_2.typeNameSymbol] });
    }
    for (const [key, value] of Object.entries(input)) {
        // Treat undefined fields and missing fields the same.
        // Generally tree does not require explicit undefined values at runtime despite some of the schema aware type checking currently requiring it.
        if (value !== undefined) {
            const childContent = extractFactoryContent(value, visitProxies !== undefined
                ? {
                    path: {
                        parent: visitProxies.path,
                        parentField: (0, index_js_3.brand)(key),
                        parentIndex: 0,
                    },
                    onVisitProxy: visitProxies?.onVisitProxy,
                }
                : undefined);
            output[key] = childContent;
        }
    }
    return output;
}
//# sourceMappingURL=proxies.js.map